##############################################################################
#
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
#
# Copyright (c) Digital Creations.  All rights reserved.
#
# This license has been certified as Open Source(tm).
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
#
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
#
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
#
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
#
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
#
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
#
#
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
#
##############################################################################

"""XMLRPC Method Product

This product provides support for external methods, which allow
domain-specific customization of web environments.
"""
__version__='$Revision: 1.6 $'[11:-2]
import xmlrpclib, httplib
from base64 import encodestring
from Globals import Persistent, DTMLFile, MessageDialog, HTML
import OFS.SimpleItem, Acquisition
from string import split, join, lower, replace
import AccessControl.Role, sys, os, stat, traceback
from OFS.SimpleItem import pretty_tb
from App.Management import Navigation
from cgi import escape
from urllib import splithost, splitpasswd, splittype, splituser
from threading import Thread

manage_addXMLRPCMethodForm=DTMLFile('dtml/methodAdd', globals())

def manage_addXMLRPCMethod(self, id, title, remoteurl, function, timeout, REQUEST=None):
    """Add an external method to a folder

    Un addition to the standard object-creation arguments,
    'id' and title, the following arguments are defined:

        method -- The name of the remote method. This can be a
          an ordinary Python function, or a bound method.

        remoteurl -- The URL of the object defining the function
          definition.

        For example, if the remoteurl is: 'ACMEWidgets.foo', then an
        attempt will first be made to use the file
        'lib/python/Products/ACMEWidgets/Extensions/foo.py'. If this
        failes, then the file 'Extensions/ACMEWidgets.foo.py' will be
        used.
    """
    id=str(id)
    title=str(title)
    remoteurl=str(remoteurl)
    function=str(function)
    timeout=float(timeout)

    i=XMLRPCMethod(id,title,remoteurl,function,timeout)
    self._setObject(id,i)
    return self.manage_main(self,REQUEST)

class XMLRPCMethod(OFS.SimpleItem.Item, Persistent, Acquisition.Explicit,
                     AccessControl.Role.RoleManager, Navigation):
    """Web-callable functions that encapsulate external python functions.

    """

    meta_type='XMLRPC Method'

    ZopeTime=Acquisition.Acquired
    HelpSys=Acquisition.Acquired

    manage_options=(
        (
        {'label':'Properties', 'action':'manage_main',
         'help':('XMLRPCMethod','XMLRPC-Method_Properties.stx')},
        {'label':'Test', 'action':'',
         'help':('XMLRPCMethod','XMLRPC-Method_Try-It.stx')},
        )
        +OFS.SimpleItem.Item.manage_options
        +AccessControl.Role.RoleManager.manage_options
        )

    __ac_permissions__=(
        ('View management screens', ('manage_main',)),
        ('Change XMLRPC Methods', ('manage_edit',)),
        ('View', ('__call__','')),
        )

    def __init__(self, id, title, remoteurl, function,timeout):
        self.id=id
        self.manage_edit(title, remoteurl, function,timeout)

    def __setstate__(self,state):
	XMLRPCMethod.inheritedAttribute('__setstate__')(self, state)
	if not hasattr(self,'_timeout'):
	    self._timeout = 5.0

    manage_main=DTMLFile('dtml/methodEdit', globals())

    def manage_edit(self, title, remoteurl, function, timeout, REQUEST=None):
        """Change the external method

        See the description of manage_addXMLRPCMethod for a
        description of the arguments 'remoteurl' and 'function'.
        """
        title=str(title)
        remoteurl=str(remoteurl)
        function=str(function)
        timeout=float(timeout)

        self.title=title
	if remoteurl:
	    urltype, url = splittype(remoteurl)
	if urltype != "http":
	    raise IOError, "Only http scheme supported"

	host, selector = splithost(url)

	if not host: raise IOError, ('http error', 'No host given')

	self._remoteurl = remoteurl
        self._function = function
        self._timeout = timeout
        if REQUEST:
            message="XMLRPC Method Created."
            return self.manage_main(self,REQUEST,manage_tabs_message=message)

    def __call__(self, *args, **kw):
        """ Call an XMLRPC Method
        """


	__traceback_info__=args, kw

	ut = RPCThread(self._remoteurl, self._function, args=args, kwargs=kw)
	ut.start()
	ut.join(self._timeout)
	return ut.getresult()

    def function(self): return self._function
    def remoteurl(self): return self._remoteurl
    def timeout(self): return self._timeout

class RPCThread(Thread):

    def __init__(self,remoteurl,function,args=(),kwargs={}):
        Thread.__init__(self)
	self.args=args
	self.kwargs=kwargs
	self.result = None
	self.rpcerror = None
	self.remoteurl = remoteurl
	self.function = function

    def getresult(self):
	if self.rpcerror:
	    raise IOError, escape(self.rpcerror)
	else:
	    return self.result

    def run(self):
	urltype, url = splittype(self.remoteurl)
	if urltype != "http":
	    self.rpcerror = "Only http scheme supported"
	    raise IOError, "Only http scheme supported"

	host, selector = splithost(url)

	if not host: raise IOError, ('http error', 'No host given')

	user_passwd, host = splituser(host)

	url = "http://"+ host + selector

	if user_passwd:
	    server = xmlrpclib.Server(url, BasicAuthTransport(user_passwd))
	else:
	    server = xmlrpclib.Server(url)
	f = server.__getattr__(self.function)
	try: res = apply(f,self.args,self.kwargs)
	except (xmlrpclib.Fault,xmlrpclib.ProtocolError,xmlrpclib.ResponseError), v:
	    self.rpcerror = str(v)
#	    raise IOError, escape(self.rpcerror)
	else:    
	    self.result = res


class BasicAuthTransport(xmlrpclib.Transport):
    def __init__(self, userpassword=None):
	self.userpassword=userpassword

    def request(self, host, handler, request_body):
	""" issue XML-RPC request """

	h = httplib.HTTP(host)
	h.putrequest("POST", handler)

	# required by HTTP/1.1
	h.putheader("Host", host)

	# required by XML-RPC
	h.putheader("User-Agent", self.user_agent)
	h.putheader("Content-Type", "text/xml")
	h.putheader("Content-Length", str(len(request_body)))

	# basic auth
	if self.userpassword is not None:
	    h.putheader("AUTHORIZATION", "Basic %s" % replace(
		    encodestring("%s" % (self.userpassword)),
		    "\012", ""))
	h.endheaders()

	if request_body:
	    h.send(request_body)

	errcode, errmsg, headers = h.getreply()

	if errcode != 200:
	    raise xmlrpclib.ProtocolError(
		host + handler,
		errcode, errmsg,
		headers
		)

	return self.parse_response(h.getfile()) 

